/* File:        $RCSfile: ProcessUtils.java,v $
 * Revision:    $Revision: 1.4 $
 * Author:      $Author: kfc $
 * Date:        $Date: 2007/02/28 16:16:26 $
 *
 * The Netarchive Suite - Software to harvest and preserve websites
 * Copyright 2004-2007 Det Kongelige Bibliotek and Statsbiblioteket, Denmark
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package dk.netarkivet.common.utils;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;

import dk.netarkivet.common.exceptions.IOFailure;

/**
 * Various utilities for running processes -- not exactly Java's forte.
 *
 */

public class ProcessUtils {
    /** Runs an external process that takes no input, discarding its output.
     *
     * @param environment An environment to run the process in (may be null)
     * @param programAndArgs The program and its arguments.
     * @return The return code of the process.
     */

    public static int runProcess(String[] environment, String... programAndArgs) {
        try {
            Process p = Runtime.getRuntime().exec(programAndArgs, environment);
            discardProcessOutput(p.getInputStream());
            discardProcessOutput(p.getErrorStream());
            while (true) {
                try {
                    return p.waitFor();
                } catch (InterruptedException e) {
                    // Ignoring interruptions, we just want to try waiting
                    // again.
                }
            }
        } catch (IOException e) {
            throw new IOFailure("Failure while running " + programAndArgs, e);
        }
    }

    /** Runs an external process that takes no input, discarding its output.
     * This is a convenience wrapper for runProcess(environment, programAndArgs)
     * @param programAndArgs The program to run and its arguments
     * @return The return code of the process.
     */
    public static int runProcess(String... programAndArgs) {
        return runProcess(null, programAndArgs);
    }

    /**
     * Read the output from a process. Due to oddities in the Process
     * handling, this has to be done char by char. This method just implements
     * a consumer thread to eat the output of a process and so prevent
     * blocking.
     *
     * @param inputStream
     *            A stream to read up to end of file. This stream is closed at
     */
    public static void discardProcessOutput(final InputStream inputStream) {
        new Thread() {
            public void run() {
                try {
                    InputStream reader = null;
                    try {
                        reader = new BufferedInputStream(inputStream);
                        int c;
                        while ((c = reader.read()) != -1) {
                        }
                    } finally {
                        if (reader != null) {
                            reader.close();
                        }
                    }
                } catch (IOException e) {
                    throw new IOFailure("Couldn't read output from process.", e);
                }
            }
        }.start();
    }
}
