/* $RCSfile: JmxHostsDeploymentBuilder.java,v $
* $Revision: 1.7 $
* $Date: 2007/02/02 11:51:14 $
* $Author: kfc $
*
* The Netarchive Suite - Software to harvest and preserve websites
* Copyright 2004-2007 Det Kongelige Bibliotek and Statsbiblioteket, Denmark
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

package dk.netarkivet.deploy;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.netarkivet.common.exceptions.ArgumentNotValid;

/**
 * Generates an add-on for a settings.xml, containing a list of JMX-hosts
 * e.g.:
 * TODO perhaps include the JmxUsername ('monitorRole') in outputXML
 * <deploy>
    <jmxMonitorRolePassword>DetErIkkeVoresSkyld</jmxMonitorRolePassword>
    <numberOfHosts>13</numberOfHosts>
    <host1>
         <name>kb-prod-adm-001.kb.dk</name>
         <jmxport>8100</jmxport>
         <jmxport>8101</jmxport>
    </host1> ..

    <host13>
         <name>kb-prod-adm-001.kb.dk</name>
         <jmxport>8100</jmxport>
         <jmxport>8101</jmxport>
    </host13>
    </deploy>
 */
public class JmxHostsDeploymentBuilder implements DeploymentBuilder {
    /** This contains the end result of this builder. */
    private final StringBuilder result;
    /** The general JMX password for the deployment. */
    private String jmxPassword;
    /** a map containing list of JMXports for each server. */
    private final Map<String,List<Integer>> hostConnectionMap =
        new HashMap<String, List<Integer>>();
    /** The location of the current host. */
    private String currentLocation;

    /**
     * Register the factory used for establishing connections to remote servers.
     * @param result the StringBuilder to write to
     */
    public JmxHostsDeploymentBuilder(StringBuilder result) {
        ArgumentNotValid.checkNotNull(result, "StringBuilder result");
        this.result = result;
    }

    /**
     * Configures the common password (for the 'monitorRole' user)
     * to use for establishing connections to all remote applications.
     */
    public void setJmxPassword(String password) {
        ArgumentNotValid.checkNotNullOrEmpty(password, "String password");
        this.jmxPassword = password;
    }

    /**
     * Returns an object that registers information about the JMXport
     * and RMIPort for each remote application
     * on the given host.
     * @return an HostBuilder object that registers information
     * about the JMXport and RMIPort
     * for each remote application on the given host.
     */
    public HostBuilder newHostBuilder() {
        return new HostBuilder() {
            private String server;
            public void setName(String serverName) {
                server = serverName;
            }
            public void addJmxPort(int jmxPortNo) {
                ArgumentNotValid.checkPositive(jmxPortNo, "int jmxPortNo");
                List<Integer> entriesForServer;
                if (hostConnectionMap.containsKey(server)) {
                    entriesForServer = hostConnectionMap.get(server);
                } else {
                    entriesForServer = new ArrayList<Integer>();
                }
                entriesForServer.add(new Integer(jmxPortNo));
                // Update the list of hostEntries
                hostConnectionMap.put(server, entriesForServer);
            }

            /**
             * Nothing needs to be done here! The action is postponed to
             * JmxProxyDeploymentBuilder.done().
             */
            public void done() {
            }
        };
    }

    /* (non-Javadoc)
     * @see dk.netarkivet.deploy.DeploymentBuilder#setLocation(java.lang.String)
     */
    public void setLocation(String location) {
        ArgumentNotValid.checkNotNullOrEmpty(location, "String location");
        this.currentLocation = location;
    }

    /*
     * (non-Javadoc)
     *
     * @see dk.netarkivet.deploy.DeploymentBuilder#done()
     */
    public void done() {
        // We have now finished collecting the information.
        // Now the real work begins.
        // 1. Check, that the JMX password is set. We need it now.
        ArgumentNotValid.checkNotNullOrEmpty(jmxPassword,
                "String jmxPassword");
        result.append("<deploy>\n");
        result.append("<jmxMonitorRolePassword>");
        result.append(jmxPassword);
        result.append("</jmxMonitorRolePassword>\n");
        result.append("<numberOfHosts>");
        int hosts = hostConnectionMap.keySet().size();
        result.append(hosts);
        result.append("</numberOfHosts>\n");

        // No need here to distinguish between hosts from different locations,
        // so location information is discarded in this builder.
        
        int i = 0;
        for (String hostname: hostConnectionMap.keySet()) {
            i++;
            result.append("<host" + i + ">\n");
            result.append("<name>" + hostname + "</name>\n");
            for (Integer jmxPort : hostConnectionMap.get(hostname)) {
                result.append("<jmxport>");
                result.append(jmxPort.longValue());
                result.append("</jmxport>\n");
            }
            result.append("</host" + i + ">\n");
        }

        result.append("</deploy>\n");
    }

}
