/* File:     $RCSfile: HeritrixLauncher.java,v $
 * Revision: $Revision: 1.7 $
 * Author:   $Author: kfc $
 * Date:     $Date: 2007/02/02 11:51:12 $
 *
 * The Netarchive Suite - Software to harvest and preserve websites
 * Copyright 2004-2007 Det Kongelige Bibliotek and Statsbiblioteket, Denmark
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package dk.netarkivet.harvester.harvesting;

import javax.management.InvalidAttributeValueException;
import java.io.File;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.archive.crawler.event.CrawlStatusListener;
import org.archive.crawler.framework.CrawlController;
import org.archive.crawler.framework.exceptions.InitializationException;
import org.archive.crawler.settings.XMLSettingsHandler;
import org.dom4j.Document;
import org.dom4j.Node;

import dk.netarkivet.common.Constants;
import dk.netarkivet.common.exceptions.ArgumentNotValid;
import dk.netarkivet.common.exceptions.IOFailure;
import dk.netarkivet.common.utils.Settings;
import dk.netarkivet.common.utils.XmlUtils;

/**
 * A HeritrixLauncher object wraps around an instance of the web crawler
 * Heritrix. The object is constructed with the necessary information to do a
 * crawl. The crawl is performed when doOneCrawl() is called. doOneCrawl()
 * monitors progress and returns when the crawl is finished or must be stopped
 * because it has stalled.
 */
public class HeritrixLauncher {
    /**
     * Class encapsulating placement of various files.
     */
    private HeritrixFiles files;

    /**
     * The CrawlController used.
     */
    private CrawlController crawlController;

    /**
     * The period to wait in milliseconds before checking if Heritrix has done
     * anything.
     */
    private final static int WAIT_PERIOD = 20000;

    //Attributes regarding deduplication.

    /**
     * Xpath for the deduplicator node in order.xml documents
     */
    static final String DEDUPLICATOR_XPATH =
            "/crawl-order/controller/map[@name='write-processors']"
            + "/newObject[@name='DeDuplicator']";

    /**
     * Xpath for the deduplicator index directory node in order.xml documents
     */
    static final String DEDUPLICATOR_INDEX_LOCATION_XPATH
            = DEDUPLICATOR_XPATH + "/string[@name='index-location']";

    /**
     * Xpath for the boolean telling if the deduplicator is enabled in
     * order.xml documents
     */
    static final String DEDUPLICATOR_ENABLED = DEDUPLICATOR_XPATH
                                               + "/boolean[@name='enabled']";

    /**
     * The class logger.
     */
    final Log log = LogFactory.getLog(getClass());

    /**
     * Has Heritrix finished crawling, yet. This field is set in the method
     * SimpleCrawlStatusListener.crawlEnded()
     */
    private AtomicBoolean crawlIsEnded = new AtomicBoolean(false);

    /** Number of milliseconds in a second. */
    private static final int MILLIS_PER_SECOND = 1000;
    /**
     * How long to wait before aborting a request from a webserver.
     */
    private static long timeOutInMillisReceivedData =
            Long.parseLong(Settings.get(
                    Settings.CRAWLER_TIMEOUT_NON_RESPONDING))
            * MILLIS_PER_SECOND;

    /**
     * How long to wait without any activity before aborting the harvest.
     */
    private static long timeOutInMillis =
            Long.parseLong(Settings.get(
                    Settings.INACTIVITY_TIMEOUT_IN_SECS)) * MILLIS_PER_SECOND;
    /** Xpath for the 'disk-path' in the order.xml */ 
    private static final String DISK_PATH_XPATH = "//crawl-order/controller"
                                                  + "/string[@name='disk-path']";
    /** Xpath for the arcfile 'prefix' in the order.xml */
    private static final String ARCFILE_PREFIX_XPATH =
            "//crawl-order/controller"
            + "/map[@name='write-processors']"
            + "/newObject/string[@name='prefix']";
    /** Xpath for the ARCs dir in the order.xml */
    private static final String ARCSDIR_XPATH = "//crawl-order/controller"
                                                + "/map[@name='write-processors']"
                                                + "/newObject/stringList[@name='path']/string";
    /** Xpath for the 'seedsfile' in the order.xml */
    private static final String SEEDS_FILE_XPATH = "//crawl-order/controller"
                                                   + "/newObject[@name='scope']"
                                                   + "/string[@name='seedsfile']";

    /**
     * Xpath for the http "user-agent" header field in order.xml-documents
     */
    public static final String USER_AGENT_XPATH = "//crawl-order/controller"
                                                  + "/map[@name='http-headers']"
                                                  + "/string[@name='user-agent']";
    /**
     * Xpath for the http "from" header field in order.xml-documents
     */
    public static final String FROM_XPATH = "//crawl-order/controller"
                                            + "/map[@name='http-headers']"
                                            + "/string[@name='from']";

    /**
     * Private HeritrixLaucher constructor. Sets up the HeritrixLauncher from
     * the given order file and seedsfile.
     *
     * @param files Object encapsulating location of Heritrix crawldir and
     *              configuration files.
     * @throws ArgumentNotValid If either seedsfile or orderfile does not
     *                          exist.
     */
    private HeritrixLauncher(HeritrixFiles files)
            throws ArgumentNotValid {
        if (!files.getOrderXmlFile().isFile()) {
            throw new ArgumentNotValid(
                    "File '" + files.getOrderXmlFile().getName()
                    + "' must exist in order for Heritrix to run. "
                    + "This filepath does not refer to existing file: "
                    + files.getOrderXmlFile().getAbsolutePath());
        }
        if (!files.getSeedsTxtFile().isFile()) {
            throw new ArgumentNotValid(
                    "File '" + files.getSeedsTxtFile().getName()
                    + "' must exist in order for Heritrix to run. "
                    + "This filepath does not refer to existing file: "
                    + files.getSeedsTxtFile().getAbsolutePath());
        }
        this.files = files;
        crawlController = new CrawlController();
    }

    /**
     * Get instance of this class.
     *
     * @param files Object encapsulating location of Heritrix crawldir and
     *              configuration files
     * @return HeritrixLauncher object
     * @throws ArgumentNotValid If either order.xml or seeds.txt does not exist
     */
    public static HeritrixLauncher getInstance(HeritrixFiles files)
            throws ArgumentNotValid {
        return new HeritrixLauncher(files);
    }

    /**
     * This method launches heritrix in the following way: 1. copies the
     * orderfile and the seedsfile to current working directory. 2. sets up the
     * newly created copy of the orderfile 3. starts the crawler 4. stops
     * crawler (Either when heritrix has finished crawling, or when heritrix is
     * forcefully stopped due to inactivity).
     *
     * The exit from the while-loop depends on Heritrix calling the crawlEnded()
     * method, when the crawling is finished. This method is called from the
     * HarvestControllerServer.onDoOneCrawl() method.
     *
     * @throws IOFailure - if the order.xml is invalid if unable to initialize
     *                   Heritrix CrawlController if Heritrix process
     *                   interrupted
     */
    public void doCrawl() throws IOFailure {
        setupOrderfile();

        try {
            // Initialize Heritrix settings according to the order.xml
            XMLSettingsHandler handler =
                    new XMLSettingsHandler(files.getOrderXmlFile());
            handler.initialize();
            crawlController.initialize(handler);

            // Start crawler
            SimpleCrawlStatusListener listener =
                    new SimpleCrawlStatusListener();
            crawlController.addCrawlStatusListener(listener);
            log.debug("Starting crawl..");
            crawlController.requestCrawlStart();
            if (crawlController.atFinish()) {
                crawlController.beginCrawlStop();
            } else {
                doCrawlLoop();
            }
        } catch (InvalidAttributeValueException e) {
            throw new IOFailure("Not valid order.xml-file", e);
        } catch (InitializationException e) {
            throw new IOFailure(
                    "Could not initialize Heritrix CrawlController", e);
        }
        log.debug("Heritrix is finished crawling...");
    }

    /**
     * Monitors the crawling performed by Heritrix. Regularly checks whether any
     * progress is made. If no progress has been made for too long, the crawl is
     * ended.
     */
    private void doCrawlLoop() {
        long lastNonZeroActiveQueuesTime = System.currentTimeMillis();
        long lastTimeReceivedData = System.currentTimeMillis();
        //TODO: Use below value!
        //long harvestedSinceLastTime = getTotalBytesWritten();
        /* Note: the crawlIsEnded is set to true inside
             the crawlEnded method, executed by Heritrix at the very end of a crawl.
        */
        while (!crawlIsEnded.get()) {
            log.info(getProgressStats());
            if (getCurrentProcessedKBPerSec() > 0) {
                lastTimeReceivedData = System.currentTimeMillis();
            }
            if (crawlController.getActiveToeCount() > 0) {
                lastNonZeroActiveQueuesTime = System.currentTimeMillis();
            }
            if ((lastNonZeroActiveQueuesTime + timeOutInMillis
                 < System.currentTimeMillis())
                || (lastTimeReceivedData + timeOutInMillisReceivedData
                    < System.currentTimeMillis())) {
                log.warn("Aborting crawl because of inactivity. "
                            + "lastNonZeroActiveQueuesTime:"
                            + lastNonZeroActiveQueuesTime
                            + "; timeOutInMillis:" + timeOutInMillis
                            + "; lastTimeReceivedData:" + lastTimeReceivedData
                            + "; timeOutInMillisReceivedData:"
                            + timeOutInMillisReceivedData
                            + "; URLs in queue:"
                            + crawlController.getFrontier().queuedUriCount());
                crawlController.requestCrawlStop(
                        "Aborting because of inactivity");
            }
            //Optimization: don't wait if ended since beginning of the loop
            if (!crawlIsEnded.get()) {
                try {
                    /* Wait for heritrix to do something
                     * WAIT_PERIOD is the interval between checks whether we have
                     * passed timeouts. Note that timeouts are defined in the settings,
                     * while WAIT_PERIOD (being less relevant to the user) is defined in
                     * this class.
                     */
                    synchronized (this) {
                        wait(WAIT_PERIOD);
                    }
                } catch (InterruptedException e) {
                    log.trace("Waiting thread awoken: " + e.getMessage());
                }
            }
        } // end of while (!crawlIsEnded)

    }

    /**
     * Make Heritrix progress statistics (String containing max 2 lines separated by
     * newline). If crawlController.getStatistics() == null,
     * the string "No statistics available" is returned.
     *
     * @return the progress statistics
     */
    private String getProgressStats() {
        if (crawlController.getStatistics() == null) {
            return "No statistics available";
        }
        String stats =
                crawlController.getStatistics().progressStatisticsLegend();
        stats += "\n"
                 + crawlController.getStatistics().getProgressStatisticsLine();
        return stats;
    }

    /**
     * Get the current value of currentProcessedKBPerSec. Uses member field
     * 'crawlController'. Returns 0, if theCrawlController.getStatistics() is
     * null
     *
     * @return the current value of currentProcessedKBPerSec.
     */
    private int getCurrentProcessedKBPerSec() {
        if (crawlController.getStatistics() == null) {
            return 0;
        }
        return crawlController.getStatistics().currentProcessedKBPerSec();
    }

    /**
     * This method prepares the orderfile used by the Heritrix crawler.
     *
     * 1. alters the orderfile in the following-way: (overriding whatever is in
     * the orderfile)
     *
     *   a. sets the disk-path to the outputdir specified in HeritrixFiles
     *
     *   b. sets the seedsfile to the seedsfile specified in HeritrixFiles.
     *
     *   c. sets the prefix of the arcfiles to unique prefix defined in
     *      HeritrixFiles
     *
     *   d. checks that the arcs-file dir is 'arcs' - to ensure that we know
     *      where the arc-files are when crawl finishes
     *
     *   e. if deduplication is enabled, sets the node pointing to index
     *      directory for deduplication (see 3)
     *
     * 2. saves the orderfile back to disk
     *
     * 3. if deduplication is enabled in order.xml; fetches lucene index over
     *    crawl.logs for jobs we use for deduplication from index server, and
     *    writes it to directory
     *
     * @throws IOFailure - When the orderfile could not be saved to disk When a
     *                   specific node is not found in the XML-document When the
     *                   SAXReader cannot parse the XML
     */
    public void setupOrderfile() throws IOFailure {
        Document doc = XmlUtils.getXmlDoc(files.getOrderXmlFile());
        XmlUtils.setNode(doc, DISK_PATH_XPATH,
                         files.getCrawlDir().getAbsolutePath());

        XmlUtils.setNode(doc, ARCFILE_PREFIX_XPATH, files.getArcFilePrefix());

        XmlUtils.setNode(doc, SEEDS_FILE_XPATH,
                         files.getSeedsTxtFile().getAbsolutePath());
        XmlUtils.setNode(doc, ARCSDIR_XPATH, Constants.ARCDIRECTORY_NAME);

        if (isDeduplicationEnabled(doc)) {
            XmlUtils.setNode(doc, DEDUPLICATOR_INDEX_LOCATION_XPATH,
                             files.getIndexDir().getAbsolutePath());
        }

        //System.out.println(doc.asXML());
        files.writeOrderXml(doc);
    }

    /**
     * Return true if the given order.xml file has deduplication enabled.
     *
     * @param doc An order.xml document
     * @return True if Deduplicator is enabled.
     */
    public static boolean isDeduplicationEnabled(Document doc) {
        Node xpath_node = doc.selectSingleNode(DEDUPLICATOR_ENABLED);
        return xpath_node != null &&
               xpath_node.getText().trim().equals("true");

    }

    /**
     * Class for handling callbacks from Heritrix. Except for logging, all that
     * happens is that - the constructor sets the value of crawlIsEnded to false
     * - crawlEnded() callback method sets the value of crawlIsEnded to true
     * crawlIsEnded is checked in the doCrawlLoop() method above. Note that
     * callbacks occur are performed by a "foreign" thread, initiated by
     * Heritrix. (That is the reason that crawlIsEnded is an AtomicBoolean).
     *
     * @see CrawlStatusListener
     */
    private class SimpleCrawlStatusListener implements CrawlStatusListener {

        /**
         * Sets the value of crawlIsEnded to false
         */
        public SimpleCrawlStatusListener() {
            HeritrixLauncher.this.crawlIsEnded.set(false);
        }

        /**
         * Fired by the crawler, when the crawl has started.
         *
         * @param s Message to attach
         * @see org.archive.crawler.event .CrawlStatusListener#crawlStarted(java.lang.String)
         */
        public void crawlStarted(String s) {
            log.debug("Crawl started: " + s);
        }

        /**
         * Fired by the crawler when the crawl is about to end (no reports
         * written yet....toe threads might still be running....).
         *
         * @param s Message to attach
         * @see org.archive.crawler.event .CrawlStatusListener#crawlEnding(java.lang.String)
         */
        public void crawlEnding(String s) {
            log.debug("Crawl ending: " + s);
        }

        /**
         * Sets the value of crawlIsEnded to true
         *
         * Fired by the crawler when the crawl is ended.
         *
         * @param s Message to attach
         * @see org.archive.crawler.event .CrawlStatusListener#crawlEnded(java.lang.String)
         */
        public void crawlEnded(String s) {
            log.info("Crawl ended: " + s);
            HeritrixLauncher.this.crawlIsEnded.set(true);
        }

        /**
         * Fired by the crawler when the crawl is about to pause (toe threads
         * might still be running.....).
         *
         * @param s Message to attach
         * @see org.archive.crawler.event .CrawlStatusListener#crawlPausing(java.lang.String)
         */
        public void crawlPausing(String s) {
            log.debug("Crawl pausing: " + s);
        }

        /**
         * Fired by the crawler when the crawl is paused.
         *
         * @param s Message to attach
         * @see org.archive.crawler.event .CrawlStatusListener#crawlPaused(java.lang.String)
         */
        public void crawlPaused(String s) {
            log.debug("Crawl paused: " + s);
        }

        /**
         * Fired by the crawler when the crawl is resuming.
         *
         * @param s Message to attach
         * @see org.archive.crawler.event .CrawlStatusListener#crawlResuming(java.lang.String)
         */
        public void crawlResuming(String s) {
            log.debug("Crawl resuming: " + s);
        }

        /**
         * Called by CrawlController when checkpointing.  Allows checkpointing
         * of local data.
         *
         * @param checkpointDir Checkpoint dir. Write checkpoint state here.
         * @throws Exception A fatal exception. Any exceptions that are let out
         *                   of this checkpoint are assumed fatal and terminate
         *                   further checkpoint processing.
         * @see org.archive.crawler.event .CrawlStatusListener#crawlCheckpoint(java.io.File)
         */
        public void crawlCheckpoint(File checkpointDir) throws Exception {
            log.debug("Crawl checkpoint to " + checkpointDir);
            // TODO: Should we store some of the parent class' information here?
        }
    }


}

