/* File:        $RCSfile: JMXUtils.java,v $
 * Revision:    $Revision: 1.7 $
 * Author:      $Author: kfc $
 * Date:        $Date: 2007/02/28 16:16:27 $
 *
 * The Netarchive Suite - Software to harvest and preserve websites
 * Copyright 2004-2007 Det Kongelige Bibliotek and Statsbiblioteket, Denmark
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package dk.netarkivet.monitor.jmx;

import javax.management.MBeanServerConnection;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.HashMap;
import java.util.Map;

import com.sun.jndi.rmi.registry.RegistryContextFactory;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import dk.netarkivet.common.exceptions.ArgumentNotValid;
import dk.netarkivet.common.exceptions.IOFailure;
import dk.netarkivet.common.exceptions.UnknownID;
import dk.netarkivet.common.utils.Settings;

/**
 * Various JMX-related utility functions that have nowhere better to live.
 *
 */
public class JMXUtils {
    public static Log log = LogFactory.getLog(Settings.class.getName());

    /** The system property that Java uses to get an initial context
     * for JNDI. This must be set for RMI connections to work.
     */
    private static final String JNDI_INITIAL_CONTEXT_PROPERTY =
        "java.naming.factory.initial";

    /**
     * If no initial JNDI context has been configured,
     * configures the system to use Sun's standard one.
     * This is necessary for RMI connections to work.
     */
    private static void ensureJndiInitialContext() {

       if (System.getProperty(JNDI_INITIAL_CONTEXT_PROPERTY) == null) {
           System.setProperty(
                   JNDI_INITIAL_CONTEXT_PROPERTY,
                   RegistryContextFactory.class.getCanonicalName());
           log.info("Set property '" + JNDI_INITIAL_CONTEXT_PROPERTY + "' to: "
                   + RegistryContextFactory.class.getCanonicalName());
       } else {
           log.debug("Property '" + JNDI_INITIAL_CONTEXT_PROPERTY
                   + "' is set to: "
                   + System.getProperty(JNDI_INITIAL_CONTEXT_PROPERTY));
       }
    }

    /**
     * Constructs the same service URL that JConsole does
     * on the basis of a server name, a JMX port number,
     * and a RMI port number.
     *
     * Example URL:
     * service:jmx:rmi://0.0.0.0:9999/jndi/rmi://0.0.0.0:1099/JMXConnector
     * where RMI port number = 9999, JMX port number = 1099
     * server = 0.0.0.0 aka localhost(?).
     *
     * @param server The server that should be connected to using
     * the constructed URL.
     * @param jmxPort The number of the JMX port that should be connected to
     * using the constructed URL
     * @param rmiPort The number of the RMI port that should be connected to
     * using the constructed URL
     * @return the constructed URL.
     */
    public static JMXServiceURL getUrl(String server,
            int jmxPort, int rmiPort) {
        String url =
            "service:jmx:rmi://" + server + ":" + rmiPort + "/jndi/rmi://"
            + server
            + ":"
            + jmxPort
            + "/jmxrmi";
        log.trace("Created url for JMX-connections: " + url);
        try {
            return new JMXServiceURL(url);
        } catch (MalformedURLException e) {
            throw new UnknownID(
                    "Could not create new JMXServiceURL from "
                    + url, e);
        }
    }

    /**
     * Returns a connection to a remote MbeanServer defined by
     * the given arguments.
     * @param server the remote servername
     * @param jmxPort the remote jmx-port
     * @param rmiPort the remote rmi-port
     * @param userName the username
     * @param password the password
     * @return a MBeanServerConnection to a remote MbeanServer defined by
     * the given arguments.
     */
    public static MBeanServerConnection getMBeanServerConnection(
            String server, int jmxPort,
            int rmiPort, String userName, String password) {
        ArgumentNotValid.checkNotNullOrEmpty(server, "String server");
        ArgumentNotValid.checkNotNegative(jmxPort, "int jmxPort");
        ArgumentNotValid.checkNotNegative(rmiPort, "int rmiPort");
        ArgumentNotValid.checkNotNullOrEmpty(userName, "String userName");
        ArgumentNotValid.checkNotNullOrEmpty(password, "String password");

        JMXServiceURL jmxServiceUrl = getUrl(server, jmxPort, rmiPort);
        Map<String,String[]> credentials =
            packageCredentials(userName, password);
        return getMBeanServerConnection(jmxServiceUrl, credentials);
    }

    /**
     * Connects to the given (url-specified) service point,
     * sending the given credentials as login.
     * @param url The JMX service url of some JVM on some machine.
     * @param credentials a map with (at least) one entry, mapping
     * "jmx.remote.credentials" to a String array of length 2.
     * Its first item should be the user name.
     * Its second item should be the password.
     * @return An MBeanServerConnection representing the
     * connected session.
     */
    public static MBeanServerConnection getMBeanServerConnection(
            JMXServiceURL url,
            Map<String,String[]> credentials) {
        try {
            ensureJndiInitialContext();
            return JMXConnectorFactory.connect(url, credentials)
                .getMBeanServerConnection();
        } catch (IOException e) {
            throw new IOFailure("Could not connect to "
                    + url.toString() , e);
        }
    }

    /**
     * Packages credentials as an environment for JMX connections.
     * This packaging has the same form that JConsole uses:
     * a one-entry Map, the mapping of "jmx.remote.credentials" being
     * an array containing the user name and the password.
     * @param userName The user to login as
     * @param password The password to use for that user
     * @return the packaged credentials
     */
    public static Map<String,String[]> packageCredentials(
            String userName, String password) {
        Map<String,String[]> credentials = new HashMap<String,String[]>(1);
        credentials.put("jmx.remote.credentials",
                new String[]{userName, password});
        return credentials;
    }
}
