/* File:        $RCSfile: DelegatingController.java,v $
 * Revision:    $Revision: 1.15 $
 * Author:      $Author: lc $
 * Date:        $Date: 2007/05/16 14:45:45 $
 *
 * The Netarchive Suite - Software to harvest and preserve websites
 * Copyright 2004-2007 Det Kongelige Bibliotek and Statsbiblioteket, Denmark
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package dk.netarkivet.viewerproxy;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import dk.netarkivet.common.distribute.indexserver.JobIndexCache;
import dk.netarkivet.common.exceptions.ArgumentNotValid;
import dk.netarkivet.common.utils.I18n;
import dk.netarkivet.common.utils.StringUtils;

/**
 * Control of viewer proxy. Delegates URL-methods to a missing URL recorder, and
 * cdx control to a CDXCache instance combined with an ARCArchiveAccess
 * instance.
 *
 */

public class DelegatingController implements Controller {
    /**
     * The missing URL recorder which handles missing URL collection.
     */
    private MissingURIRecorder mur;
    /**
     * The CDX cache which generates a CDX from a list of jobs on changeIndex
     * command.
     */
    private JobIndexCache cc;
    /**
     * The ARCArchiveAccess instance to receive new cdx on changeIndex command.
     */
    private ARCArchiveAccess aaa;
    /**
     * Remembers the last joblist used for setting index for status purposes.
     */
    private Set<Long> jobSet;
    /**
     * Remembers the label of the index for status purposes.
     */
    private String indexLabel;
    /**
     * Internationalisation context.
     */
    private final static I18n I18N = new I18n(Constants.TRANSLATIONS_BUNDLE);

    /**
     * Initialise a controller with the relevant instances to control
     *
     * @param mur The missing URL recorder which handles missing URL
     *            collection.
     * @param cc  The CDX cache which generates an index from a list of jobs on
     *            changeIndex command.
     * @param aaa The ARCArchiveAccess instance to receive new cdx on
     *            changeIndex command.
     * @throws ArgumentNotValid if any argument is null.
     */
    public DelegatingController(MissingURIRecorder mur, JobIndexCache cc,
                      ARCArchiveAccess aaa) {
        ArgumentNotValid.checkNotNull(mur, "MissingURIRecorder mur");
        ArgumentNotValid.checkNotNull(cc, "CDXCache cc");
        ArgumentNotValid.checkNotNull(aaa, "ARCArchiveAccess aaa");
        this.mur = mur;
        this.cc = cc;
        this.aaa = aaa;
    }

    /**
     * Start collecting missing URLs
     */
    public void startRecordingURIs() {
        mur.startRecordingURIs();
    }

    /**
     * Stop collecting missing URLs
     */
    public void stopRecordingURIs() {
        mur.stopRecordingURIs();
    }

    /**
     * Clear list of missing URLs
     */
    public void clearRecordedURIs() {
        mur.clearRecordedURIs();
    }

    /**
     * Get list of missing URLs.
     *
     * @return The list of missing URLs.
     */
    public Set<URI> getRecordedURIs() {
        return mur.getRecordedURIs();
    }

    /**
     * Change index to use an index based on a list of jobs. Note: Does not
     * check arguments. This is a task for the mediated classes,
     * ArcArchiveAccess and CDXCache.
     *
     * @param jobSet List of jobs to get an index for.
     * @param label The label this index should be known as
     */
    public void changeIndex(Set<Long> jobSet, String label) {
        aaa.setIndex(cc.getIndex(jobSet));
        this.jobSet = jobSet;
        this.indexLabel = label;
    }

    /** Get a human readable status of the viewer proxy.
     *
     * @return A human readable status string.
     * @param locale The locale used to generate the string
     */
    public String getStatus(Locale locale) {
        ArgumentNotValid.checkNotNull(locale, "locale");
        StringBuilder status = new StringBuilder();
        if (mur.isRecordingURIs()) {
            status.append(I18N.getString(locale, "currently.collecting.urls"));
        } else {
            status.append(I18N.getString(locale,
                    "currently.not.collecting.urls"));
        }
        status.append('\n');
        status.append(I18N.getString(locale, "current.list.contains.0.urls",
                mur.getRecordedURIs().size()));
        status.append('\n');
        if (jobSet == null) {
            status.append(I18N.getString(locale, "no.index.set"));
        } else {
            List<Long> jobList = new ArrayList<Long>(jobSet);
            Collections.sort(jobList);
            status.append(I18N.getString(locale, "index.0.built.on.jobs.1",
                                         indexLabel,
                                         StringUtils.conjoin(jobList, ", ")));
        }
        return status.toString();
    }
}
